@@ -2,9 +2,13 @@
 import asyncio
 import logging
 import socket
-from contextlib import asynccontextmanager, contextmanager, suppress
+from contextlib import contextmanager, suppress
+try:
+    from contextlib import asynccontextmanager
+except ImportError:
+    from async_generator import asynccontextmanager
 import paho.mqtt.client as mqtt
-from .error import MqttError, MqttCodeError
+from .error import MqttError, MqttCodeError, MqttConnectError
 
 
 MQTT_LOGGER = logging.getLogger('mqtt')
@@ -12,7 +16,9 @@ MQTT_LOGGER.setLevel(logging.WARNING)
 
 
 class Client:
-    def __init__(self, hostname, port=1883, *, username=None, password=None, logger=None):
+    def __init__(self, hostname, port=1883, *, username=None, password=None,
+                 logger=None, client_id=None, tls_context=None, protocol=None,
+                 will=None):
         self._hostname = hostname
         self._port = port
         self._loop = asyncio.get_event_loop()
@@ -22,7 +28,10 @@ class Client:
         self._pending_calls_threshold = 10
         self._misc_task = None
 
-        self._client = mqtt.Client()
+        if protocol is None:
+            protocol = mqtt.MQTTv311
+
+        self._client = mqtt.Client(client_id=client_id, protocol=protocol)
         self._client.on_connect = self._on_connect
         self._client.on_disconnect = self._on_disconnect
         self._client.on_subscribe = self._on_subscribe
@@ -42,14 +51,51 @@ class Client:
         if username is not None and password is not None:
             self._client.username_pw_set(username=username, password=password)
 
-    async def connect(self):
-        self._client.connect(self._hostname, self._port, 60)
-        await self._connected
-        self._client.socket().setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)
+        if tls_context is not None:
+            self._client.tls_set_context(tls_context)
 
-    async def disconnect(self):
-        self._client.disconnect()
-        await self._disconnected
+        if will is not None:
+            self._client.will_set(
+                will.topic,
+                will.payload,
+                will.qos,
+                will.retain,
+                will.properties
+            )
+
+    @property
+    def id(self):
+        """Return the client ID.
+
+        Note that paho-mqtt stores the client ID as `bytes` internally.
+        We assume that the client ID is a UTF8-encoded string and decode
+        it first.
+        """
+        return self._client._client_id.decode()
+
+    async def connect(self, *, timeout=10):
+        try:
+            self._client.connect(self._hostname, self._port, 60)
+            # paho.mqttClient.socket() return non-None after the call to
+            # connect.
+            self._client.socket().setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2048)
+        # paho.mqtt.Client.connect may raise one of several exceptions.
+        # We convert all of them to the common MqttError for user convenience.
+        # See: https://github.com/eclipse/paho.mqtt.python/blob/v1.5.0/src/paho/mqtt/client.py#L1770
+        except (socket.error, OSError, mqtt.WebsocketConnectionError) as error:
+            raise MqttError(str(error))
+        await self._wait_for(self._connected, timeout=timeout)
+
+    async def disconnect(self, *, timeout=10):
+        rc = self._client.disconnect()
+        # Early out on error
+        if rc != mqtt.MQTT_ERR_SUCCESS:
+            raise MqttCodeError(rc, "Could not disconnect")
+        # Wait for acknowledgement
+        await self._wait_for(self._disconnected, timeout=timeout)
+
+    async def force_disconnect(self):
+        self._disconnected.set_result(None)
 
     async def subscribe(self, *args, timeout=10, **kwargs):
         result, mid = self._client.subscribe(*args, **kwargs)
@@ -98,7 +144,7 @@ class Client:
         Example use:
             async with client.filtered_messages('floors/+/humidity') as messages:
                 async for message in messages:
-                    print(f'Humidity reading: {message.decode()}')
+                    print(f'Humidity reading: {message.payload.decode()}')
         """
         cb, generator = self._cb_and_generator(log_context=f'topic_filter="{topic_filter}"',
                                                queue_maxsize=queue_maxsize)
@@ -140,7 +186,28 @@ class Client:
         async def _message_generator():
             # Forward all messages from the queue
             while True:
-                yield await messages.get()
+                # Wait until we either:
+                #  1. Receive a message
+                #  2. Disconnect from the broker
+                get = self._loop.create_task(messages.get())
+                try:
+                    done, _ = await asyncio.wait(
+                        (get, self._disconnected),
+                        return_when=asyncio.FIRST_COMPLETED
+                    )
+                except asyncio.CancelledError:
+                    # If the asyncio.wait is cancelled, we must make sure
+                    # to also cancel the underlying tasks.
+                    get.cancel()
+                    raise
+                if get in done:
+                    # We received a message. Return the result.
+                    yield get.result()
+                else:
+                    # We got disconnected from the broker. Cancel the "get" task.
+                    get.cancel()
+                    # Stop the generator with the following exception
+                    raise MqttError("Disconnected during message iteration")
         return _put_in_queue, _message_generator()
 
     async def _wait_for(self, *args, **kwargs):
@@ -170,25 +237,42 @@ class Client:
             # network error) we still need to remove the item from the dict.
             self._pending_calls.pop(mid, None)
                 
-    def _on_connect(self, client, userdata, flags, rc):
+    def _on_connect(self, client, userdata, flags, rc, properties=None):
         if rc == mqtt.CONNACK_ACCEPTED:
             self._connected.set_result(rc)
         else:
-            self._connected.set_exception(MqttCodeError(rc, 'Could not connect'))
+            self._connected.set_exception(MqttConnectError(rc))
 
-    def _on_disconnect(self, client, userdata, rc):
+    def _on_disconnect(self, client, userdata, rc, properties=None):
+        # Return early if the disconnect is already acknowledged.
+        # Sometimes (e.g., due to timeouts), paho-mqtt calls _on_disconnect
+        # twice. We return early to avoid setting self._disconnected twice
+        # (as it raises an asyncio.InvalidStateError).
+        if self._disconnected.done():
+            return
+        # Return early if we are not connected yet. This avoids calling
+        # `_disconnected.set_exception` with an exception that will never
+        # be retrieved (since `__aexit__` won't get called if `__aenter__`
+        # fails). In turn, this avoids asyncio debug messages like the
+        # following:
+        #
+        #   "[asyncio] Future exception was never retrieved"
+        #
+        # See also: https://docs.python.org/3/library/asyncio-dev.html#detect-never-retrieved-exceptions
+        if not self._connected.done():
+            return
         if rc == mqtt.MQTT_ERR_SUCCESS:
             self._disconnected.set_result(rc)
         else:
-            self._disconnected.set_exception(MqttCodeError(rc, 'Could not disconnect'))
+            self._disconnected.set_exception(MqttCodeError(rc, 'Unexpected disconnect'))
 
-    def _on_subscribe(self, client, userdata, mid, granted_qos):
+    def _on_subscribe(self, client, userdata, mid, granted_qos, properties=None):
         try:
             self._pending_calls.pop(mid).set_result(granted_qos)
         except KeyError:
             MQTT_LOGGER.error(f'Unexpected message ID "{mid}" in on_subscribe callback')
 
-    def _on_unsubscribe(self, client, userdata, mid):
+    def _on_unsubscribe(self, client, userdata, mid, properties=None, reasonCodes=None):
         try:
             self._pending_calls.pop(mid).set()
         except KeyError:
@@ -206,11 +290,11 @@ class Client:
     def _on_socket_open(self, client, userdata, sock):
         def cb():
             client.loop_read()
-        self._loop.add_reader(sock, cb)
+        self._loop.add_reader(sock.fileno(), cb)
         self._misc_task = self._loop.create_task(self._misc_loop())
 
     def _on_socket_close(self, client, userdata, sock):
-        self._loop.remove_reader(sock)
+        self._loop.remove_reader(sock.fileno())
         with suppress(asyncio.CancelledError):
             self._misc_task.cancel()
 
@@ -231,6 +315,31 @@ class Client:
         await self.connect()
         return self
 
-    async def __aexit__(self, *args, **kwargs):
+    async def __aexit__(self, exc_type, exc, tb):
         """Disconnect from the broker."""
-        await self.disconnect()
+        # Early out if already disconnected...
+        if self._disconnected.done():
+            disc_exc = self._disconnected.exception()
+            if disc_exc is None:
+                # ...by raising the error that caused the disconnect
+                raise disc_exc
+            # ...by returning since the disconnect was intentional
+            return
+        # Try to gracefully disconnect from the broker
+        try:
+            await self.disconnect()
+        except MqttError as error:
+            # We tried to be graceful. Now there is no mercy.
+            MQTT_LOGGER.warning(f'Could not gracefully disconnect due to "{error}". Forcing disconnection.')
+            await self.force_disconnect()
+
+
+# TODO: This should be a (frozen) dataclass (from Python 3.7)
+# when we drop Python 3.6 support
+class Will:
+    def __init__(self, topic, payload=None, qos=0, retain=False, properties=None):
+        self.topic = topic
+        self.payload = payload
+        self.qos = qos
+        self.retain = retain
+        self.properties = properties
